__all__ = [
    'load_open_mmlab_models',
    'transfer_weight',
    'transfer_weights',
    'save_checkpoint',
    'load_checkpoint',
]

from typing import Any, Dict, Optional

import torch
import torch.nn as nn
import torch.optim as optim
import torch.optim.lr_scheduler as lr_scheduler

from ._extensions import Module, get_logger, getattr_recur
from .configs import Config
from .registries import Registry


def load_open_mmlab_models(
    registry: Registry,
    config: str,
    config_options: Optional[Config] = None,
    ckpt: Optional[str] = None,
) -> Module:
    import mmcv
    import mmcv.runner

    config_dict = mmcv.Config.fromfile(config).model
    model = (
        registry.build(config_dict) if config_options is None else
        registry.build(config_options, default_args=config_dict)
    )
    if ckpt is not None:
        mmcv.runner.load_checkpoint(model, ckpt, map_location='cpu')
        model._is_init = True
    return model


def transfer_weight(target: Module, source: Module) -> None:
    state_dict = source.state_dict()
    missing_keys, unexpected_keys = target.load_state_dict(
        state_dict,
        strict=False,
    )
    if missing_keys:
        get_logger().warning('missing_keys:', missing_keys)
    if unexpected_keys:
        get_logger().warning('unexpected_keys:', unexpected_keys)
    target._is_init = True


def transfer_weights(models, weight_prefixes: Dict[str, str]) -> None:
    for target_prefix, source_prefix in weight_prefixes.items():
        target = getattr_recur(models, target_prefix)
        source = getattr_recur(models, source_prefix)
        transfer_weight(target, source)


def save_checkpoint(
    model: nn.Module,
    f: str,
    *,
    optimizer: Optional[optim.Optimizer] = None,
    scheduler: Optional[lr_scheduler._LRScheduler] = None,
    meta: Optional[Config] = None,
    **kwargs,
) -> None:
    # TODO: state dict to cpu
    checkpoint: Dict[str, Any] = dict()
    checkpoint['model'] = model.state_dict()
    if optimizer is not None:
        checkpoint['optimizer'] = optimizer.state_dict()
    if scheduler is not None:
        checkpoint['scheduler'] = scheduler.state_dict()
    if meta is not None:
        checkpoint['meta'] = meta
    torch.save(checkpoint, f, **kwargs)


def load_checkpoint(
    model: nn.Module,
    f: str,
    *,
    optimizer: Optional[optim.Optimizer] = None,
    scheduler: Optional[lr_scheduler._LRScheduler] = None,
    **kwargs,
) -> Optional[Config]:
    checkpoint: Dict[str, Any] = torch.load(f, **kwargs)
    model.load_state_dict(checkpoint['model'])
    if optimizer is not None:
        optimizer.load_state_dict(checkpoint['optimizer'])
    if scheduler is not None:
        scheduler.load_state_dict(checkpoint['lr_scheduler'])
    return checkpoint.get('meta')



import hashlib
import random
from contextlib import contextmanager
from typing import Generator, Optional, cast

import numpy as np
import torch
import torch.backends.cudnn as cudnn
import torch.distributed as dist

from ..base import (
    get_iter,
    get_logger,
    get_rank,
    get_world_size,
    iter_initialized,
)


def _randint(high: int = 2**30) -> int:
    seed = np.random.randint(high)
    if get_world_size() <= 1:
        return seed
    tensor = torch.tensor(seed if get_rank() == 0 else 0, dtype=torch.int)
    dist.broadcast(tensor, src=0)
    return cast(int, tensor.item())


def init_seed(
    seed=None,
    deterministic: Optional[bool] = None,
) -> int:
    if seed is None:
        seed = _randint()
    elif isinstance(seed, int):
        pass
    else:
        if not isinstance(seed, bytes):
            seed = str(seed).encode()
        seed = hashlib.blake2b(seed, digest_size=4).hexdigest()
        seed = int(seed, 16)
    seed %= 2**30

    if iter_initialized():
        seed += get_iter()

    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

    if deterministic is not None:
        cudnn.deterministic = deterministic
        cudnn.benchmark = not deterministic

    return seed


@contextmanager
def set_seed_temp(
    seed=None,
    deterministic: Optional[bool] = None,
) -> Generator[None, None, None]:
    if torch.cuda.device_count() > 1:
        get_logger().warn(
            "Seeding is not recommended for multi-GPU training.",
        )
    random_state = random.getstate()
    np_state = np.random.get_state()
    torch_state = torch.get_rng_state()
    cuda_state = torch.cuda.get_rng_state_all()
    prev_detereministic = cudnn.deterministic
    prev_benchmark = cudnn.benchmark
    init_seed(seed, deterministic)

    yield

    random.setstate(random_state)
    np.random.set_state(np_state)
    torch.set_rng_state(torch_state)
    torch.cuda.set_rng_state_all(cuda_state)
    cudnn.deterministic = prev_detereministic
    cudnn.benchmark = prev_benchmark



__all__ = [
    'BaseAdapt',
    'ADAPTS',
]

import inspect
import itertools

import einops.layers.torch
import torch.nn as nn

from ..base import STEPS, Module, Registry


class BaseAdapt(Module):
    pass


ADAPTS: Registry[nn.Module] = Registry('adapts', parent=STEPS, base=nn.Module)
for _, class_ in itertools.chain(
    inspect.getmembers(nn, inspect.isclass),
    inspect.getmembers(einops.layers.torch, inspect.isclass),
):
    if issubclass(class_, nn.Module):
        ADAPTS.register(class_)

try:
    import mmcv.cnn

    for k, v in itertools.chain(
        mmcv.cnn.CONV_LAYERS.module_dict.items(),
        mmcv.cnn.PLUGIN_LAYERS.module_dict.items(),
    ):
        ADAPTS.register(v, name=f'mmcv_{k}')
except Exception:
    pass



__all__ = [
    'ChainedScheduler',
]

import math
from typing import Iterable

from .base import SCHEDULERS, BaseScheduler, SchedulerConfig


@SCHEDULERS.register_module()
class ChainedScheduler(BaseScheduler):

    def __init__(
        self,
        schedulers: Iterable[SchedulerConfig],
        value: float = 1.0,
    ) -> None:
        self._schedulers = tuple(map(SCHEDULERS.build, schedulers))
        self._value = value

    @property
    def value(self) -> float:
        return math.prod(self._schedulers) * self._value



# Modified from
# https://github.com/open-mmlab/mmcv/blob/v1.6.1/mmcv/utils/config.py
#
# Copyright (c) OpenMMLab. All rights reserved.

__all__ = [
    'Config',
]

import argparse
import difflib
import pathlib
import tempfile
import webbrowser
from functools import reduce
from typing import (
    Any,
    Dict,
    Iterable,
    Mapping,
    NoReturn,
    Optional,
    Type,
    TypeVar,
)

import addict
import yapf.yapflib.yapf_api as yapf

T = TypeVar('T', bound='Config')

BASE = '_base_'
DELETE = '_delete_'


class Config(addict.Dict):

    def __missing__(self, name) -> NoReturn:
        raise KeyError(name)

    def __getattr__(self, name: str):
        try:
            return super().__getattr__(name)
        except KeyError as e:
            raise AttributeError(e)

    @classmethod
    def merge(cls, a, b):
        if not isinstance(b, Mapping):
            return b

        b = cls(b)
        if isinstance(a, list) and all(isinstance(x, int) for x in b):
            a = list(a)
            while len(a) in b:
                a.append(b.pop(len(a)))
            for i in sorted(b):
                a[i] = (
                    cls.merge(a[i], b[i])
                    if isinstance(a[i], Mapping) else b[i]
                )
            return a

        if b.pop(DELETE, False):
            return b
        if not isinstance(a, Mapping):
            return b

        a = cls(a)
        for k in b:
            a[k] = cls.merge(a[k], b[k]) if k in a else b[k]
        return a

    @classmethod
    def loads(cls: Type[T], s: str, globals: Optional[dict] = None) -> T:
        if globals is None:
            globals = dict()
        globals.setdefault('__name__', '__main__')

        config: Dict[str, Any] = dict()
        exec(s, globals, config)
        return cls(config)

    @classmethod
    def load(cls: Type[T], file) -> T:
        file = pathlib.Path(file)
        file = file.resolve()

        config = cls.loads(file.read_text())
        configs = [
            cls.load(file.parent / base) for base in config.pop(BASE, [])
        ]
        configs.append(config)
        return reduce(cls.merge, configs)

    @classmethod
    def diff(
        cls: Type[T],
        a: T,
        b: T,
        mode: str = 'text',
    ) -> str:
        a_ = a.dumps().split('\n')
        b_ = b.dumps().split('\n')
        if mode == 'text':
            return '\n'.join(difflib.Differ().compare(a_, b_))
        if mode == 'html':
            return difflib.HtmlDiff().make_file(a_, b_)
        raise ValueError(f"Invalid mode {mode}.")

    def dumps(self) -> str:

        def format(obj) -> str:
            contents: Iterable[str]
            if isinstance(obj, dict):
                if all(isinstance(k, str) and k.isidentifier() for k in obj):
                    contents = [k + '=' + format(v) for k, v in obj.items()]
                    delimiters = ('dict(', ')')
                else:
                    contents = [
                        format(k) + ': ' + format(v) for k, v in obj.items()
                    ]
                    delimiters = ('{', '}')
            elif isinstance(obj, list):
                contents = map(format, obj)
                delimiters = ('[', ']')
            elif isinstance(obj, tuple):
                contents = map(format, obj)
                delimiters = ('(', ')')
            elif isinstance(obj, set):
                contents = map(format, obj)
                delimiters = ('{', '}')
            else:
                return repr(obj)
            contents = sorted(contents)
            if len(obj) != 1:
                contents.append('')
            return delimiters[0] + ','.join(contents) + delimiters[1]

        assert all(isinstance(k, str) for k in self)
        code = '\n'.join(k + ' = ' + format(self[k]) for k in sorted(self))
        code, _ = yapf.FormatCode(code, verify=True)
        return code

    def dump(self, file) -> None:
        file = pathlib.Path(file)
        file = file.resolve()
        file.write_text(self.dumps())


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Compare Configs")
    parser.add_argument('a')
    parser.add_argument('b')
    parser.add_argument('--out')
    args = parser.parse_args()

    a = Config(args.a)
    b = Config(args.b)

    if args.out is None:
        diff_mode = 'text'
    elif args.out.endswith('.txt'):
        diff_mode = 'text'
    elif args.out.endswith('.html'):
        diff_mode = 'html'
    elif args.out == 'browser':
        diff_mode = 'html'
    else:
        raise ValueError(f"Unknown output mode: {args.out}.")

    diff = Config.diff(a, b, diff_mode)
    if args.out is None:
        print(diff)
    elif args.out == 'browser':
        with tempfile.NamedTemporaryFile(
            suffix='.html',
            delete=False,
        ) as html_file:
            html_file.write(diff.encode('utf-8'))
            webbrowser.open('file://' + html_file.name)
    else:
        with open(args.out, 'w') as f:
            f.write(diff)



from ._extensions import *
from ._patches import *
from .bboxes import *
from .checkpoints import *
from .configs import *
from .debug import *
from .iters import *
from .registries import *
from .status import *
from .workflows import *



import itertools
from enum import Enum, auto
from functools import partial
from typing import Any, Iterable, Iterator, Optional, TypeVar

import torch.nn as nn
from torch.nn.modules.batchnorm import _BatchNorm

from ..base import getattr_recur

DEFAULT_STOCHASTIC_MODULE_TYPES = [_BatchNorm, nn.Dropout]

try:
    from timm.models.layers import DropPath

    DEFAULT_STOCHASTIC_MODULE_TYPES.append(DropPath)
except Exception:
    pass


class NoGradMode(Enum):
    ALL = auto()
    PARTIAL = auto()
    NONE = auto()

    def no_grad(self, model: nn.Module, **kwargs) -> nn.Module:
        if self == NoGradMode.ALL:
            return model.requires_grad_(False)
        if self == NoGradMode.NONE:
            return model

        modules: Iterator[nn.Module]
        if self == NoGradMode.PARTIAL:
            modules = map(
                partial(getattr_recur, model),
                kwargs['module_names'],
            )
        else:
            raise NotImplementedError(f"{self} is not implemented")

        for module in modules:
            module.requires_grad_(False)
        return model


class EvalMode(Enum):
    ALL = auto()
    DETERMINISTIC_AND_PARTIAL = auto()
    DETERMINISTIC = auto()
    PARTIAL = auto()
    PARTIALLY_DETERMINISTIC = auto()
    NONE = auto()

    def eval(
        self,
        model: nn.Module,
        **kwargs,
    ) -> nn.Module:
        if self == EvalMode.ALL:
            return model.eval()
        if self == EvalMode.NONE:
            return model

        stochastic_module_types: Iterable[Any] = kwargs.get(
            'stochastic_module_types',
            DEFAULT_STOCHASTIC_MODULE_TYPES,
        )

        def stochastic(module: nn.Module) -> bool:
            return any(
                isinstance(module, smt) for smt in stochastic_module_types
            )

        modules: Iterator[nn.Module]
        if self == EvalMode.DETERMINISTIC:
            modules = filter(stochastic, model.modules())
        elif self == EvalMode.PARTIAL:
            modules = map(
                partial(getattr_recur, model),
                kwargs['module_names'],
            )
        elif self == EvalMode.PARTIALLY_DETERMINISTIC:
            modules = map(
                partial(getattr_recur, model),
                kwargs['module_names'],
            )
            modules = filter(stochastic, modules)
        elif self == EvalMode.DETERMINISTIC_AND_PARTIAL:
            named_modules = map(
                partial(getattr_recur, model),
                kwargs['module_names'],
            )
            stochastic_modules = filter(stochastic, model.modules())
            modules = itertools.chain(named_modules, stochastic_modules)
        else:
            raise NotImplementedError(f'EvalMode {self} not implemented')

        for module in modules:
            module.eval()
        return model


def freeze_model(model: nn.Module, no_grad=..., eval_=...) -> nn.Module:
    if no_grad is ...:
        no_grad = dict(mode='ALL')
    if no_grad is not None:
        no_grad = dict(no_grad)
        no_grad_mode_name: str = no_grad.pop('mode')
        no_grad_mode = NoGradMode[no_grad_mode_name.upper()]
        no_grad_mode.no_grad(model, **no_grad)
    if eval_ is ...:
        eval_ = dict(mode='ALL')
    if eval_ is not None:
        eval_ = dict(eval_)
        eval_mode_name: str = eval_.pop('mode')
        eval_mode = EvalMode[eval_mode_name.upper()]
        eval_mode.eval(model, **eval_)
    return model


T = TypeVar('T', bound='FrozenMixin')


class FrozenMixin(nn.Module):

    def __init__(
        self,
        *args,
        freeze_cfg: Optional[dict] = None,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._freeze_cfg = freeze_cfg or dict()
        freeze_model(self, **self._freeze_cfg)

    def requires_grad_(self: T, requires_grad: bool = True) -> T:
        result = super().requires_grad_(requires_grad)
        freeze_model(self, **self._freeze_cfg)
        return result

    def train(self: T, mode: bool = True) -> T:
        result = super().train(mode)
        freeze_model(self, **self._freeze_cfg)
        return result



__all__ = [
    'DebugMode',
]

import os

from ._extensions import get_logger


class DebugMode:

    def __init__(self) -> None:
        self._logger = get_logger()

    def __set_name__(self, owner, name: str) -> None:
        self._name = name
        self._logger.debug(f"Debug mode {name} is {self.__get__(None)}")

    def __get__(self, obj, objtype=None) -> bool:
        return bool(os.getenv(self._name))

    def __set__(self, obj, value: bool) -> None:
        if value:
            os.environ[self._name] = '1'
        else:
            os.environ.pop(self._name)
        self._logger.debug(
            f"Debug mode {self._name} is set to {self.__get__(None)}"
        )
